/******************************************************************//**
 * Zerocat Chipflasher --- Flash free firmware to BIOS chips,
 *                         kick the Manageability Engine.
 *
 * Copyright (C) 2015, 2016, 2017, 2018 Kai Mertens <kmx@posteo.net>
 *
 * The Zerocat Chipflasher is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * The Zerocat Chipflasher is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This file is part of the Zerocat Chipflasher.
 *
 *
 * @file
 *
 */

#ifndef __CHIPSPEC_H__
  #error "chipspec.h has not been included, yet."
#endif

#ifndef __SERIAL_CODES_H__
  #error "serial-codes.h has not been included, yet."
#endif

#ifndef __FILESPEC_H__
  #error "filespec.h has not been included, yet."
#endif

#ifndef __KICK_H__
#define __KICK_H__

//define doxygen module groups
/**
 * @defgroup Motorola Motorola
 *
 * @brief
 * Objects that are needed to deal with the Motorola S-record format.
 *
 */

/**
 * @defgroup Hexdump Hexdump
 *
 * @brief
 * Objects that are needed to deal with the Hexdump format.
 *
 */

//start diagnostic compiler switches
#define ERROR_MSG__LINE_TOO_LONG
//end diagnostic compiler switches


//constants
#define HEADER_KICK         " Zerocat Chipflasher: kick " VERSION " "
#define SIZE_STR_MESSAGE    (1 << 7)
#define SIZE_STRBUF         (1 << 8)
#define SIZE_LINEBUF        (1 << 8)
#define BAUDRATE_DEFAULT    115200    //!< propeller-load default baudrate
#define NOT_APPLICABLE      -1        //!< some chipspec members may not make sense for some chips
#define POWERUP_SPI         25        //!< Give the attached system board some time to power up. Time in milliseconds.
#define POWERUP_SPILINES    25        //!< Give SPI lines some time to power up (i.e. GA-G41M-ES2L).
#define POWERUP_EXTRA       2         //!< If chip detection fails, how often shall we retry?
#define POWERUP_OFFTIME     100000    //!< Separate multiple chip detections, time in µ-seconds.
#define STACK_DEFAULT       0x40      //!< Default stack size for any extra cog function, then to be optimized.
#define STACK_BURNBUF       0x20      //!< Stack size for burn().
#define STACK_LEDSTAT       0x20      //!< Stack size for ledstat().
#define MAX_POLLING         100000000 //!< Set a timeout value for the WIP polling routine.

//modes, used by main()
#define MODE_TERMINAL       0x01
#define MODE_CONNECT        0x02

//modes, used by SPI_off()
#define MODE_SPI_OFF        0x00
#define MODE_FORCE_SPI_OFF  0x01
#define MODE_SPI_POWEROFF   0x02

//used by ledstat()
#define LOOP_RES            ((CLKFREQ / 1000000) * 1000)   //1ms
#define STATUS_CYCLE        1000   //1000ms
#define LED_PHASE           90     //90ms

//macros
#define PUTCHAR_CR          do {                                    \
                              outfifo(CARR_RET);                    \
                            } while(0)

#define PUTCHAR_NL          do {                                    \
                              outfifo(NEW_LINE);                    \
                            } while(0)

#define GET_CHAR_GROUP(n)   for(unsigned char i = 0; i < n; i++) {  \
                              c[i] = outfifo(getChar());            \
                            }

#define MSG_DONE            "...done.\r\n\n"

#define SPI_ON              SPI_ini()
#define SPI_OFF(mode)       SPI_off(mode)



enum tag_cmdstat {
  cmd_off = 0,
  cmd_on,
  cmd_CP
};

enum tag_spimode {
  mode_0 = 0,
  mode_3
};

enum tag_type_of_addrlen {
  ADDRLEN_IN_BYTES = 0,
  ADDRLEN_MOT,   //minimal lenght is 2, not 1 or zero
  ADDRLEN_IN_BITS
};

enum tag_blocksize {
  BLOCK_4K = SIZE_4K,
  BLOCK_32K = SIZE_32K,
  BLOCK_64K = SIZE_64K
};



/**
 * @ingroup Motorola
 *
 * @brief
 * Motorola S-Record Type Infos.
 *
 */
union tag_typeS {
  unsigned int mem;
  struct {
    unsigned int none:1;
    unsigned int start:1;
    unsigned int data:1;
    unsigned int linecount:1;
    unsigned int startexec:1;
    unsigned int addrlen:3;
    unsigned int num:4;
    unsigned int :0;
  } bf;
};



/**
 * @brief
 * Struct that characterizes a data line.
 */
 struct tag_linedat {
  char * buf;                   //!< Pointer to buffer, that holds the data.
  unsigned int mask;            //!< Mask for read and write pointers.
  unsigned int rd;              //!< Read pointer.
  unsigned int wr;              //!< Write pointer.
  unsigned int startaddr;       //!< Line’s startaddress.
  unsigned char payload;        //!< Number of payload bytes.
};



/**
 * @brief
 * Cog parameters for ledstat().
 *
 */
struct tag_xcog1 {
  volatile unsigned char green;     //!< Switch for green LED usage.
  volatile unsigned char yellow;    //!< Switch for yellow LED usage.
  volatile unsigned char orange;    //!< Switch for orange LED usage.
  volatile unsigned char inlevel;   //!< Current menu input level.
};



/**
 * @brief
 * Cog parameters for burn().
 *
 * @note
 * Not all members have to be declared volatile, for access is
 * controlled by run and busy.
 *
 * @warning
 * Do not declare flags as a volatile bitfield struct. It does not work.
 *
 */
struct tag_xcog0 {
  int * pcog;                           //!< Cog info pointer, used by cog_end().
  unsigned char cmd;                    //!< Command to be used for writing.
  unsigned char abits;                  //!< addrbits.
  unsigned int lineaddr;                //!< startaddress of data line.
  int psize;                            //!< Page size.
  unsigned char * pbuf;                 //!< Pointer to ringbuffer.
  unsigned char bufsize;                //!< Size of ringbuffer.
  volatile unsigned int offrd;          //!< Buffer read pointer offset.
  volatile unsigned int offwr;          //!< Buffer write pointer offset.
  volatile unsigned char queue_empty;   //!< Flag to indicate that the data queue is empty.
  volatile unsigned char rq_spioff;     //!< Flag to request SPI off.
  volatile unsigned char rq_lineaddr;   //!< Flag to request the next line address.
};



/**
 * @ingroup ChipspecData
 *
 * @brief
 * Struct that characterizes a chip.
 *
 */
struct tag_chip {
  unsigned char index;                  //!<
  struct tag_chipspec * pspec;          //!< Pointer into chipspec database.
  unsigned char addrlen_in_bits;        //!< Number of bits that are used feed an address into the SPI-bus.
  int pages;                            //!< Number of pages (on a paged chip).
};



int hex2bin (
  unsigned char a,
  unsigned char b
  );

void chip_erase(
  void
  );

void HEXDUMP_mkline (
  struct tag_linedat * linedat,
  char * linebuf_orig
  );

char HEXDUMP_rxline (
  struct tag_xcog0 * px,
  unsigned int * lines
  );

void MOT_mkline(
  struct tag_linedat * linedat,
  union tag_typeS * typeS,
  char * linebuf
  );

char MOT_rxline (
  struct tag_xcog0 * px,
  unsigned int * lines,
  int hexmode
  );

void MOT_typeS_fillup(
  union tag_typeS * typeS
  );

void chip_txfile (
  struct tag_filespec * filespec,
  unsigned char addrlen_in_bits,
  unsigned int firstloc,
  unsigned int memsize,
  unsigned char screenmode
  );

void menu_line(
  unsigned int k1,
  char * k1msg,
  unsigned int k2,
  char * k2msg,
  unsigned int k3,
  char * k3msg
  );

void cmd_DP (
  void
  );

signed char linebuf_out (
  char * linebuf
  );

unsigned char get_addrlen(
  enum tag_type_of_addrlen type_of_addrlen,
  unsigned int addr
  );

void outbits (
  const unsigned int value,
  unsigned int msbit
  );

int cmd_RDID_JEDEC (
  void
  );

char queue (
  struct tag_xcog0 * px,
  unsigned char vbin,
  unsigned int * pwr
  );

void chip_rxfile (
  struct tag_xcog0 * px,
  char screen_output
  );

signed char chip_read (
  struct tag_filespec * filespec,
  unsigned int firstloc,
  unsigned int memsize,
  unsigned int * lines,
  char * linebuf
  );

void ledstat();
void burn();

void MOT_header(
  char * header,
  char * linebuf
  );

void MOT_data (
  struct tag_linedat * linedat,
  char * linebuf
);

void MOT_summary (
  unsigned int lines,
  char * linebuf
  );

void chip_ini (
  unsigned char new_index
  );

unsigned int CPM_polling (
  void
  );

void PGM_cycle (
  enum tag_cmdstat * cmdstat,
  unsigned char odd_tracker
  );

unsigned char verify_0xff(
  unsigned int firstloc,
  unsigned int memsize
  );

inline unsigned char inb (
  void
  );

unsigned int inbits (
  unsigned int msbit
  );

void key_config(
  void
  );

void menu_options(
  void
  );

void greeting (
  );

unsigned char yes_no (
  char * question,
  unsigned char level_rst
  );

void exit_sequence (
  const char status
  );

void SPI_ini (
  void
  );

void SPI_off (
  int mode
  );

int WIP_polling (
  void
  );

int strpos(
  char * str,
  char c
  );

void cmd (
  const unsigned char cmd,
  const unsigned int value,
  unsigned char bits
  );

unsigned char cmd_RDSR (
  unsigned char regno
  );

void cmd_WRSR (
  const unsigned char regno,
  const unsigned char byte
  );

unsigned char SR_report (
  unsigned char regno
  );

void SR_intro(
  unsigned char regno
  );

void cmd_WREN (
  void
  );

void cmd_WRDI (
  void
  );

void global_sector_protect (
  unsigned char pswitch
  );

void option_readblock (
  const enum tag_blocksize blocksize
  );

void option_batchblockerase(
  const enum tag_blocksize blocksize,
  const unsigned char blockcmd
  );

int do__WIP_polling(
  void
  );

void key_polling(
  unsigned char * key_is_valid,
  unsigned char * quit
  );

void range_error(
  void
  );

void page_write (
  const unsigned int addr,
  unsigned char * buffer
  );

int plug_detect(
  void
  );

void menu(
  unsigned char mode
  );

char outfifo (
  char c
  );

int outstr (
  char * s
  );

void connected (
  void
  );

void hr(
  char c,
  int n,
  char nl
  );

void mirror (
  char * keymsg
  );

void putChar(
  char c
  );

void report_err(
  char reported_err
  );

void check_dryrun(
  void
  );


#endif
/* __KICK_H__ */
